/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
package org.greenstone.gatherer.gui.table;

import java.awt.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import javax.swing.table.*;
import org.greenstone.gatherer.Configuration;
import org.greenstone.gatherer.Dictionary;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.file.FileNode;
import org.greenstone.gatherer.gui.WarningDialog;
import org.greenstone.gatherer.msm.ElementWrapper;
import org.greenstone.gatherer.msm.Metadata;
import org.greenstone.gatherer.msm.MSMAction;
import org.greenstone.gatherer.msm.MSMEvent;
import org.greenstone.gatherer.msm.MSMListener;
import org.greenstone.gatherer.msm.MSMUtils;
import org.greenstone.gatherer.util.ArrayTools;
import org.greenstone.gatherer.util.Utility;

/** Provides the model for a GTable component, filling it with metadata values for the choosen files or folders. The model also provides several different view of this data; assigned folder metadata, assigned file metadata, all assigned metadata, unassigned metadata, and all metadata. It also differentiates between metadata that is common to all of the files or folders, and that which isn't. The building of the actual model is done on a separate thread so that the gui remains responsive, and the gui is intermitantly updated by this thread. Updating of the model is triggered by events recieved from the metadata set manager in terms of new or obsolete metadata. A new model is rebuilt whenever the user selects a different group of files or folders.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.3b
 */
public class GTableModel
    extends AbstractTableModel
    implements MSMListener {

    /** The source for the data currently shown in the table. May be any of the other three metadata lists, or the union of all of them. */
    private ArrayList current_metadata = new ArrayList();
    /** The file nodes this model is built upon. */
    private FileNode[] file_nodes = null;
    /** The table this model is being displayed in. */
    private JTable table = null;

    private ModelBuilder builder = null;

    static final private String[] COLUMN_NAMES = {"", Dictionary.get("Metadata.Element"),  Dictionary.get("Metadata.Value")};


    public GTableModel(JTable table) {
	this.table = table;
	// Register this model with the msm manager.
	if (Gatherer.c_man.ready()) {
	    Gatherer.c_man.getCollection().msm.addMSMListener(this);
	}
    }


    public GTableModel(JTable table, FileNode[] file_nodes) {
	this(table);
	this.file_nodes = file_nodes;
	if (file_nodes != null && file_nodes.length > 0) {
	    // Create model builder
	    builder = new ModelBuilder();
	    builder.start();
	}
    }


    /** Returns the number of columns in this table. */
    public int getColumnCount() {
	return COLUMN_NAMES.length;
    }


    /** Retrieves the name of the specified column. */
    public String getColumnName(int col) {
  	return COLUMN_NAMES[col];
    }


    /** Returns the number of rows in this table. */
    public int getRowCount() {
	return current_metadata.size();
    }


    /** Returns the cell value at a given row and column as an Object. */
    public Object getValueAt(int row, int col) {
	Object result = "";
	if(0 <= row && row < current_metadata.size()) {
	    Metadata m = (Metadata) current_metadata.get(row);
	    if(0 <= col && col < COLUMN_NAMES.length) {
		switch(col) {
		case 0:
		    if(!m.isFileLevel()) {
			result = m.getFile();
		    }
		    break;
		case 1:
		    result = m.getElement();
		    break;
		case 2:
		    result = m.getValue();
		    break;
		}
	    }
	}
	return result;
    }


    /* Called to retrieve the Metadata record that matches a certain row. Usually caused by the user selectiong a row in the table. It synchronized so that the model doesn't up and change while we're trying to retrieve the indicated element. */
    public synchronized Metadata getMetadataAtRow(int row) {
	if(0 <= row && row < current_metadata.size()) {
	    return (Metadata) current_metadata.get(row);
	}
	return null;
    }


    public void selectMetadataWhenBuildingComplete(Metadata metadatum)
    {
	// If there is no builder or no data then there is no point...
	if (builder == null || metadatum == null) {
	    return;
	}

	// If building is not complete, ask the builder to select the metadata when finished
	if (!builder.complete) {
	    builder.select_exact_metadata = metadatum;
	}
	// Otherwise, select the appropriate row in the table
	else {
	    for (int i = 0; i < current_metadata.size(); i++) {
		// Find the matching piece of metadata
		if (current_metadata.get(i).equals(metadatum)) {
		    // We have to put the selection task onto the AWT thread
		    SelectionTask task = new SelectionTask(i);
		    SwingUtilities.invokeLater(task);
		    return;
		}
	    }
	}
    }


    public void selectClosestMetadataWhenBuildingComplete(Metadata metadatum)
    {
	// If there is no builder or no data then there is no point...
	if (builder == null) {
	    return;
	}

	// If building is not complete, ask the builder to select the metadata when finished
	if (!builder.complete) {
	    builder.select_closest_metadata = metadatum;
	}
	// Otherwise, select the appropriate row in the table
	else {

	    // If there was no previous metadata, select the first row (if any)
	    if(metadatum == null) {
		if(current_metadata.size() > 0) {
		    // We have to put the selection task onto the AWT thread
		    SelectionTask task = new SelectionTask(0);
		    SwingUtilities.invokeLater(task);
		    task = null;
		}
		return;
	    }

	    // Find the position in the list where the piece of metadata should be
	    Comparator metadata_comparator = MSMUtils.METADATA_COMPARATOR;
	    int i = 0;
	    for (i = 0; i < current_metadata.size(); i++) {
		if (metadata_comparator.compare(metadatum, current_metadata.get(i)) <= 0) {
		    break;
		}
	    }

	    // Ensure the position is within the current metadata
	    if (i >= current_metadata.size()) {
			i = current_metadata.size() - 1;
	    }

	    // Select the same row as before, except if it would select a different element
	    ElementWrapper element = metadatum.getElement();
	    if(i < 0) {
			  Gatherer.println("Could not find suitable row to select!");
			  return;
		 }
	    Metadata newly_selected_metadatum = (Metadata) current_metadata.get(i);
	    if (!newly_selected_metadatum.getElement().equals(element)) {
			// Try one row previous
			i--;
		    if(i < 0) {
				  Gatherer.println("Could not find suitable row to select!");
				  return;
			 }

			newly_selected_metadatum = (Metadata) current_metadata.get(i);
			if (!newly_selected_metadatum.getElement().equals(element)) {
			    Gatherer.println("Could not find suitable row to select!");
			    return;
			}
	    }

	    // We have to put the selection task onto the AWT thread
	    SelectionTask task = new SelectionTask(i);
	    SwingUtilities.invokeLater(task);
	}
    }


    /** Little utility class for selecting a row in the table, on the AWT thread. */
    private class SelectionTask
	extends Thread {

	private int selected_row;

	SelectionTask(int selected_row) {
	    this.selected_row = selected_row;
	}

	public void run() {
	    Rectangle bounds = table.getCellRect(selected_row, 0, true);
	    table.scrollRectToVisible(bounds);
	    table.setRowSelectionInterval(selected_row, selected_row);
	}
    }


    /** Called whenever an element in a set is added, edited or removed from the metadata set manager. We listen for this just incase the name of one of the elements currently shown changes. After that the table will be up to date, as the element references are otherwise live. */
    public void elementChanged(MSMEvent event) {
	// Brute force solution... rebuild the table.
	rebuildModel();
    }


    /** Whenever there is a metadata change we must determine if there are any changes to our various lists of metadata, and also if there is any need to refresh the table currently being viewed. */
    public void metadataChanged(MSMEvent event) {
	FileNode file_node = event.getRecord();
	// First check whether this record is one of those that we have showing.
	if (file_nodes != null && file_nodes.length > 0 && file_node != null && ArrayTools.contains(file_nodes, file_node)) {
	    // Brute force solution... rebuild the table.
	    rebuildModel();
	}
    }


    /** Called whenever a set is added or removed from the metadata set manager. */
    public void setChanged(MSMEvent event) {
	// No, don't care. Still don't care. Couldn't care less.
    }


    /** Called when the value tree of a certain element has changed significantly but, although we display metadata values, we only care about those coming through metadataChanged() events. */
    public void valueChanged(MSMEvent event) {
	// Don't care. Tra-la-la-la-la.
    }


    private void rebuildModel()
    {
	///ystem.err.println("Rebuild model");
	Metadata select_exact_metadata = null;
	Metadata select_closest_metadata = null;
	// If there is a current builder...
	if (builder != null) {
	    // ...and it is still building...
	    if (!builder.complete) {
		select_exact_metadata = builder.select_exact_metadata;
		select_closest_metadata = builder.select_closest_metadata;
		// ...kill it...
		builder.kill();
	    }
	    // ...and wait until it dies
	    while (!builder.complete) { }
	}

	// Create model builder
	builder = new ModelBuilder();
	builder.start();
	builder.select_exact_metadata = select_exact_metadata;
	builder.select_closest_metadata = select_closest_metadata;
    }

    /** Determine if the given metadata is common to all selected file nodes given the context of the current view. */
   public boolean isCommon(Metadata entry) {
      return (file_nodes != null && entry.getCount() >= file_nodes.length);
   }

    /** Determine if the given metadata is common to all selected file nodes given the context of the current view. */
    public boolean isCommon(int row) {
	if(0 <= row && row < current_metadata.size()) {
	    Metadata entry = (Metadata) current_metadata.get(row);
	    return isCommon(entry);
	}
	return false;
    }

    public boolean isBuildingComplete() {
	return (builder == null || builder.complete);
    }

    private class ModelBuilder
	extends Thread {

	public boolean complete = false;
	public Metadata select_exact_metadata = null;
	public Metadata select_closest_metadata = null;
	private boolean had_extracted_warning = false;
	private boolean had_inherited_warning = false;
	private volatile boolean killed = false;

	public void run()
	{
	    if (!Gatherer.c_man.ready()) {
		Gatherer.println("Not ready!");
		return;
	    }

	    boolean have_encountered_file = false;

	    ///ystem.err.println("\nBuilding metadata model...");
	    Vector elements = Gatherer.c_man.getCollection().msm.getElements();
	    ArrayList known_elements = new ArrayList();
	    current_metadata.clear();

	    for (int i = 0; i < file_nodes.length && !killed; i++) {
		File current_file = file_nodes[i].getFile();
		have_encountered_file = have_encountered_file || current_file.isFile();
		///ystem.err.println("File " + i + " (of " + file_nodes.length + "): " + current_file);

		// Get the currently assigned metadata for this file
		ArrayList metadata = Gatherer.c_man.getCollection().gdm.getAllMetadata(current_file);
		for (int j = 0; j < metadata.size() && !killed; j++) {
		    Metadata metadatum = (Metadata) metadata.get(j);
		    ElementWrapper element = metadatum.getElement();
		    ///ystem.err.println("  Metadatum: " + metadatum + " (" + metadatum.getCount() + ")");

		    // Ignore hidden metadata
		    if (!element.toString().startsWith("hidden.")) {
			// Only process extracted metadata if the user has specified to
			if(!element.getNamespace().equals(Utility.EXTRACTED_METADATA_NAMESPACE) || Gatherer.config.get("general.view_extracted_metadata", Configuration.COLLECTION_SPECIFIC)) {
			    // If a piece of metadata is at folder level, display the warning
			    if (!metadatum.isFileLevel()) {
				showInheritedMetadataWarning();
			    }

			    // If the piece of metadata is in the list, increment its count
			    int index = current_metadata.indexOf(metadatum);
			    if (index >= 0) {
				Metadata existing = (Metadata) current_metadata.get(index);
				existing.inc();
			    }
			    // Otherwise add it to the list
			    else {
				metadatum.setCount(1);
				index = add(current_metadata, metadatum);
				fireTableRowsInserted(index, index);

				// Remember we have seen this element
				if (known_elements.contains(element) == false) {
				    known_elements.add(element);
				}
			    }
			}
			// Display a warning
			else {
			    showExtractedMetadataWarning();
			}
		    }
		}
	    }

	    // Make sure each element has a piece of metadata (even if blank)
	    for (int i = 0; i < elements.size() && !killed; i++) {
		ElementWrapper element = (ElementWrapper) elements.get(i);
		///ystem.err.println("Element: " + element);

		// If we haven't seen this metadata element, and conditional on the fact that if it is extracted, we have seen at least one file, then add it now
		//System.err.println("Element name = " + element);
		//System.err.println("Element extracted = " + element.isExtracted());
		//System.err.println("have_encountered_files = " + have_encountered_file);
		if (!known_elements.contains(element)) { // && (!element.isExtracted() || have_encountered_file)) {
		    Metadata metadatum = new Metadata(element);
		    metadatum.setCount(file_nodes.length);

		    // Add it to the table
		    int index = add(current_metadata, metadatum);
		    fireTableRowsInserted(index, index);
		}
	    }

	    // Sort the metadata and update the table
	    if (!killed) {
		Gatherer.g_man.metaedit_pane.validateMetadataTable();
		Collections.sort(current_metadata, MSMUtils.METADATA_COMPARATOR);
		fireTableDataChanged();
	    }

	    // Processing complete
	    complete = true;

	    // If while processing we've been asked to select a certain metadata, do it now
	    if (!killed) {
		if (select_exact_metadata != null) {
		    selectMetadataWhenBuildingComplete(select_exact_metadata);
		}
		else {
		    selectClosestMetadataWhenBuildingComplete(select_closest_metadata);
		}
	    }
	}


	public synchronized void kill()
	{
	    killed = true;
	    notify();
	}


	/** Alphabetically inserts the new_metadata in the target vector. */
	private int add(ArrayList target, Metadata new_metadata) {
	    int i = 0;
	    while(i < target.size()) {
		Metadata current = (Metadata)target.get(i);
		if(current.compareTo(new_metadata) > 0) {
		    target.add(i, new_metadata);
		    return i;
		}
		i++;
	    }
	    target.add(new_metadata);
	    return i;
	}

	private void showExtractedMetadataWarning()
	{
	    if (!had_extracted_warning) {
		had_extracted_warning = true;
		Runnable task = new Runnable() {
			public void run() {
			    WarningDialog dialog = new WarningDialog("warning.ExtractedMetadata", false);
			    dialog.display();
			    dialog.dispose();
			    dialog = null;
			}
		    };
		SwingUtilities.invokeLater(task);
	    }
	}

	private void showInheritedMetadataWarning()
	{
	    if (!had_inherited_warning) {
		had_inherited_warning = true;
		Runnable task = new Runnable() {
			public void run() {
			    WarningDialog dialog = new WarningDialog("warning.InheritedMetadata", false);
			    dialog.display();
			    dialog.dispose();
			    dialog = null;
			}
		    };
		SwingUtilities.invokeLater(task);
	    }
	}
    }
}
